////////////////////////////////////////////////////////////////////////////
//
// nrModel3DS.cpp
//
// Nate Robins, November 2001.
//
// See http://astronomy.swin.edu.au/pbourke/3dformats/3ds/
//
// 3D Studio 3DS model file format reader.
//
////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////

#include "nrModel3DS.h"
#include "nrModelMTL.h"

#include "nrArray.h"
#include "nrModel.h"
#include "nrVector2.h"
#include "nrVector3.h"

#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <string.h>


////////////////////////////////////////////////////////////////////////////
// Defines
////////////////////////////////////////////////////////////////////////////

#define LINE_SIZE        1024


////////////////////////////////////////////////////////////////////////////
// Classes
////////////////////////////////////////////////////////////////////////////

class nrChunker
{
public:

    inline nrChunker( void );
    inline ~nrChunker( void );
    
    // Open a stream (file) for chunking.
    bool Open( const char* file_name );
    
    // Close the stream (file) being chunked.
    void Close( void );
    
    // Prepare the chunker for the next chunk.
    // 
    // If the next chunk doesn't exist (end of file), false is returned,
    // otherwise true is returned.
    inline bool Next( void );
    
    // Skip the current chunk.
    inline void Skip( void );
    
    // Reads the expected chunk id from the stream.
    // 
    // If the next chunk in the stream is not of the expected id, a parse 
    // error is issued.
    inline bool Read( unsigned short id );
    
    // Reads a string from the stream (reads up to a '\0').
    // 
    // If an error occurs, a parse is issued and 0 is returned.
    inline char* ReadString( void );
    
    // Returns the number of bytes read so far.
    inline int Bytes( void ) const;

    // Return the ID of the current chunk.
    inline unsigned short ID( void ) const;
    
    // Return the length (in bytes) of the data in the current chunk.
    inline unsigned short Length( void ) const;
    
    // Return the data of the current chunk (should be delete'd).
    inline unsigned char* Data( void );
    
private:
    
    FILE* m_File;
    
    unsigned short m_ID;
    unsigned int   m_Length;
    unsigned char* m_Data;
    
    unsigned int m_Byte;
    unsigned int m_Bytes;
};

////////////////////////////////////////////////////////////////////////////
// Implementation
////////////////////////////////////////////////////////////////////////////

nrChunker::nrChunker( void )
{
    m_File = 0;

    m_ID = 0;
    m_Length = 0;
    m_Data = 0;

    m_Byte = 0;
    m_Bytes = 0;
}

////////////////////////////////////////////////////////////////////////////

nrChunker::~nrChunker( void )
{
}

////////////////////////////////////////////////////////////////////////////

bool nrChunker::Open( const char* file_name )
{
    m_File = fopen( file_name, "rb" );
    if ( m_File == 0 )
    {
        return false;
    }
    
    m_Byte = 0;
    
    fseek( m_File, 0, SEEK_END );
    m_Bytes = ftell( m_File );
    fseek( m_File, 0, SEEK_SET );
    
    m_ID = 0;
    m_Length = 0;
    
    return true;
}

////////////////////////////////////////////////////////////////////////////

void nrChunker::Close( void )
{
    fclose( m_File );
    m_File = 0;
}

////////////////////////////////////////////////////////////////////////////

int nrChunker::Bytes( void ) const
{
    return m_Byte;
}

////////////////////////////////////////////////////////////////////////////

unsigned short nrChunker::ID( void ) const
{
    return m_ID;
}

////////////////////////////////////////////////////////////////////////////

unsigned short nrChunker::Length( void ) const
{
    return m_Length;
}

////////////////////////////////////////////////////////////////////////////

unsigned char* nrChunker::Data( void )
{
    m_Data = new unsigned char[ m_Length ];
    if ( m_Data == 0 )
    {
        fprintf( stderr, "nrChunker::Data(): Out of memory.\n" );
        assert( 0 );
        return 0;
    }
    
    if ( fread( m_Data, m_Length, 1, m_File ) != 1 )
    {
        fprintf( stderr, "nrChunker::Data(): Error reading chunk data (file may be truncated).\n" );
        delete [] m_Data;
        assert( 0 );
        return 0;
    }
    
    m_Byte += m_Length;
    
    return m_Data;
}

////////////////////////////////////////////////////////////////////////////

bool nrChunker::Read( unsigned short id )
{
    assert( m_File );

    if ( Next() )
    {
        if ( m_ID == id )
        {
            return true;
        }
        else
        {
            return false;
        }
    }
    else
    {
        return false;
    }
}

////////////////////////////////////////////////////////////////////////////

char* nrChunker::ReadString( void )
{
    assert( m_File );
    nrArray< char > string( 256 );
    
    int c;
    while ( c = getc( m_File ) )
    {
        string.Add( c );
    }
    
    char* s = new char[ string.Length() + 1 ];
    
    for ( int i = 0; i < string.Length(); i++ )
    {
        s[ i ] = string[ i ];
    }
    s[ i ] = 0;
    
    m_Byte += i + 1;
    
    return s;
}

////////////////////////////////////////////////////////////////////////////

bool nrChunker::Next( void )
{
    assert( m_File );
    
    int c = getc( m_File );
    if ( feof( m_File ) )
    {
        return false;
    }
    ungetc( c, m_File );
    
    if ( fread( &( m_ID ), 2, 1, m_File ) != 1 )
    {
        fprintf( stderr, "nrChunker::Next(): Error reading chunk ID (file may be truncated).\n" );
        return false;
    }
    
    if ( fread( &( m_Length ), 4, 1, m_File ) != 1 )
    {
        fprintf( stderr, "nrChunker::Next(): Error reading chunk length (file may be truncated).\n" );
        return false;
    }
    
    // Already read 4 bytes from this chunk (ID == 2, Length == 2), so make
    // sure we've got a length of at least 4, and then subtract from the length.
    if ( m_Length < 4 )
    {
        fprintf( stderr, "nrChunker::Next(): Invalid chunk length (file may be corrupt).\n" );
        return false;
    }
    
    m_Length -= 6;
    m_Byte += 6;
    
    return true;
}

////////////////////////////////////////////////////////////////////////////

void nrChunker::Skip( void )
{
    fseek( m_File, m_Length, SEEK_CUR );
    
    m_Byte += m_Length;
}


////////////////////////////////////////////////////////////////////////////
// Implementation
////////////////////////////////////////////////////////////////////////////

nrModel3DS::nrModel3DS()
{
}

////////////////////////////////////////////////////////////////////////////

nrModel3DS::~nrModel3DS()
{
}

////////////////////////////////////////////////////////////////////////////

#define NEW( ptr, type, num ) \
    ptr = new type[ num ]; \
    \
    if ( ptr == NULL ) \
    { \
        fprintf( stderr, "nrModel3DS::NewFromFile(): Out of memory.\n" ); \
        delete m_Model; \
        return NULL; \
    }

////////////////////////////////////////////////////////////////////////////

#undef NEW

////////////////////////////////////////////////////////////////////////////

nrModel* nrModel3DS::NewFromFile( const char* file_name )
{
    nrModel3DS l;
    
    nrChunker chunker;
    
    if ( ! chunker.Open( file_name ) )
    {
        fprintf( stderr, "nrModel3DS::NewFromFile(): Unable to open file \"%s\".\n", file_name );
        return 0;
    }
    
    nrModel* m = 0;
    
    if ( chunker.Read( 0x4D4D ) )
    {
        l.Chunk_4D4D( chunker );
        m = new nrModel();
    }
    else
    {
        fprintf( stderr, "nrModel3DS::NewFromFile(): Invalid 3DS file \"%s\".\n", file_name );
    }
    
    chunker.Close();
    
    return m;
}

////////////////////////////////////////////////////////////////////////////

void nrModel3DS::Chunk_4D4D( nrChunker& chunker )
{
    int total = chunker.Bytes() + chunker.Length();
    
    while ( chunker.Bytes() < total )
    {
        chunker.Next();

        unsigned short id = chunker.ID();
        unsigned short length = chunker.Length();
        
        printf( "Chunk: 0x%04X  Length: %d bytes\n", id, length );
        
        switch ( id )
        {
        case 0x0002:    // 3DS version
            printf( "nrModel3DS::Chunk_4D4D(): Chunk 0x%04X, skipped.\n", id );
            chunker.Skip();
            break;
            
        case 0x3D3D:    // 3D Editor chunk
            Chunk_3D3D( chunker );
            break;
            
        case 0xB000:    // Keyframer chunk
            printf( "nrModel3DS::Chunk_4D4D(): Chunk 0x%04X, skipped.\n", id );
            chunker.Skip();
            break;
            
        default:
            printf( "nrModel3DS::Chunk_4D4D(): Unknown chunk 0x%04X, skipped.\n", id );
            chunker.Skip();
            break;
        }
    }
}

////////////////////////////////////////////////////////////////////////////

void nrModel3DS::Chunk_3D3D( nrChunker& chunker )
{
    int total = chunker.Bytes() + chunker.Length();
    
    while ( chunker.Bytes() < total )
    {
        chunker.Next();

        unsigned short id = chunker.ID();
        unsigned short length = chunker.Length();
        
        printf( "Chunk: 0x%04X  Length: %d bytes\n", id, length );
        
        switch ( id )
        {
        case 0x0100:    // One unit

        case 0x1100:    // Background bitmap
        case 0x1101:    // Use background bitmap

        case 0x1200:    // Background color
        case 0x1201:    // Use background color

        case 0x1300:    // Gradient colors
        case 0x1301:    // Use gradient

        case 0x1400:    // Shadow map bias
        case 0x1420:    // Shadow map size
        case 0x1450:    // Shadow map sample range

        case 0x1460:    // Raytrace bias
        case 0x1470:    // Raytrace on

        case 0x2100:    // Ambient color

        case 0x2200:    // Fog
            // 0x2210 : fog background
        case 0x2201:    // Use fog
        case 0x2210:    // Fog background

        case 0x2300:    // Distance queue
            // 0x2310 : Dim background
        case 0x2301:    // Use distance queue

        case 0x2302:    // Layered fog options
        case 0x2303:    // Use layered fog

        case 0x3D3E:    // Mesh version

        case 0x7001:    // Window settings
            // 0x7011 : Window description #2 ...
            // 0x7012 : Window description #1 ...
            // 0x7020 : Mesh windows ...
            printf( "nrModel3DS::Chunk_3D3D(): Chunk 0x%04X, skipped.\n", id );
            chunker.Skip();
            break;
            
        case 0x4000:    // Object block
            Chunk_4000( chunker );
            break;
            
        case 0xAFFF:    // Material block
            Chunk_AFFF( chunker );
            break;
            
        default:
            printf( "nrModel3DS::Chunk_3D3D(): Unknown chunk 0x%04X, skipped.\n", id );
            chunker.Skip();
            break;
        }
    }
}

////////////////////////////////////////////////////////////////////////////

void nrModel3DS::Chunk_4000( nrChunker& chunker )
{
    int total = chunker.Bytes() + chunker.Length();
    
    // First bit of a 4000 chunk is the name.
    char* name = chunker.ReadString();
    printf( "Name: %s.\n", name );
    delete [] name;
    
    while ( chunker.Bytes() < total )
    {
        chunker.Next();
        
        unsigned short id = chunker.ID();
        unsigned short length = chunker.Length();
        
        printf( "Chunk: 0x%04X  Length: %d bytes\n", id, length );
        
        switch ( id )
        {
        case 0x4010:    // Object hidden
        case 0x4012:    // Object doesn't cast
        case 0x4013:    // Matte object
        case 0x4015:    // External process on
        case 0x4017:    // Object doesn't receive shadows
            printf( "nrModel3DS::Chunk_4000(): Chunk 0x%04X, skipped.\n", id );
            chunker.Skip();
            break;
            
        case 0x4100:    // Triangular mesh
            Chunk_4100( chunker );
            break;
            
        case 0x4600:    // Light
            // 0x4610 : Spotlight
            //   0x4627 : Spot raytrace
            //   0x4630 : Light shadowed
            //   0x4641 : Spot shadow map
            //   0x4650 : Spot show cone
            //   0x4651 : Spot is rectangular
            //   0x4652 : Spot overshoot
            //   0x4653 : Spot map
            //   0x4656 : Spot roll
            //   0x4658 : Spot ray trace bias
            
            // 0x4620 : Light off
            // 0x4625 : Attenuation on
            // 0x4659 : Range start
            // 0x465A : Range end
            // 0x465B : Multiplier
        case 0x4700:    // Camera
            printf( "nrModel3DS::Chunk_4000(): Chunk 0x%04X, skipped.\n", id );
            chunker.Skip();
            break;
            
        default:
            printf( "nrModel3DS::Chunk_4000(): Unknown chunk 0x%04X, skipped.\n", id );
            chunker.Skip();
            break;
        }
    }
}

////////////////////////////////////////////////////////////////////////////

void nrModel3DS::Chunk_4100( nrChunker& chunker )
{
    int total = chunker.Bytes() + chunker.Length();
    
    while ( chunker.Bytes() < total )
    {
        chunker.Next();

        unsigned short id = chunker.ID();
        unsigned short length = chunker.Length();
        
        printf( "Chunk: 0x%04X  Length: %d bytes\n", id, length );
        
        switch ( id )
        {
        case 0x4110:    // Vertices list

        case 0x4120:    // Faces description
            // 0x4130 : Faces material list

        case 0x4140:    // Mapping coordinates list
            // 0x4150 : Smoothing group list

        case 0x4160:    // Local coordinate system
        case 0x4165:    // Object color in editor

        case 0x4181:    // External process name
        case 0x4182:    // External process parameters
            printf( "nrModel3DS::Chunk_4100(): Chunk 0x%04X, skipped.\n", id );
            chunker.Skip();
            break;
            
        default:
            printf( "nrModel3DS::Chunk_4100(): Unknown chunk 0x%04X, skipped.\n", id );
            chunker.Skip();
            break;
        }
    }
}

////////////////////////////////////////////////////////////////////////////

//void nrModel3DS::Chunk_4110( nrChunker& chunker )
//{
//
//}

////////////////////////////////////////////////////////////////////////////

void nrModel3DS::Chunk_AFFF( nrChunker& chunker )
{
    int total = chunker.Bytes() + chunker.Length();
    
    while ( chunker.Bytes() < total )
    {
        chunker.Next();

        unsigned short id = chunker.ID();
        unsigned short length = chunker.Length();
        
        printf( "Chunk: 0x%04X  Length: %d bytes\n", id, length );
        
        switch ( id )
        {
        case 0xA000:    // Material name
            
        case 0xA010:    // Ambient color
        case 0xA020:    // Diffuse color
        case 0xA030:    // Specular color
            
        case 0xA040:    // Shininess percent
        case 0xA041:    // Shininess strength percent
            
        case 0xA050:    // Transparency percent
        case 0xA052:    // Transparency falloff percent
        case 0xA053:    // Reflection blur percent
            
        case 0xA081:    // 2 sided
        case 0xA083:    // Add trans
        case 0xA084:    // Self illum
        case 0xA085:    // Wire frame on
        case 0xA087:    // Wire thickness
        case 0xA088:    // Face map
        case 0xA08A:    // In tranc
        case 0xA08C:    // Soften
        case 0xA08E:    // Wire in units
            
        case 0xA100:    // Render type
            
        case 0xA240:    // Transparency falloff percent present
        case 0xA250:    // Reflection blur percent present
        case 0xA252:    // Bump map present (true percent)
            
        case 0xA200:    // Texture map 1
        case 0xA33A:    // Texture map 2
        case 0xA210:    // Opacity map
        case 0xA230:    // Bump map
        case 0xA33C:    // Shininess map
        case 0xA204:    // Specular map
        case 0xA33D:    // Self illum. map
        case 0xA220:    // Reflection map
        case 0xA33E:    // Mask for texture map 1
        case 0xA340:    // Mask for texture map 2
        case 0xA342:    // Mask for opacity map
        case 0xA344:    // Mask for bump map
        case 0xA346:    // Mask for shininess map
        case 0xA348:    // Mask for specular map
        case 0xA34A:    // Mask for self illum. map
        case 0xA34C:    // Mask for reflection map
            // Sub-chunks for all maps:
            //  0xA300 : Mapping filename
            //  0xA351 : Mapping parameters
            //  0xA353 : Blur percent
            //  0xA354 : V scale
            //  0xA356 : U scale
            //  0xA358 : U offset
            //  0xA35A : V offset
            //  0xA35C : Rotation angle
            //  0xA360 : RGB Luma/Alpha tint 1
            //  0xA362 : RGB Luma/Alpha tint 2
            //  0xA364 : RGB tint R
            //  0xA366 : RGB tint G
            //  0xA368 : RGB tint B
            printf( "nrModel3DS::Chunk_AFFF(): Chunk 0x%04X, skipped.\n", id );
            chunker.Skip();
            break;
            
        default:
            printf( "nrModel3DS::Chunk_AFFF(): Unknown chunk 0x%04X, skipped.\n", id );
            chunker.Skip();
            break;
        }
    }
}

////////////////////////////////////////////////////////////////////////////
