////////////////////////////////////////////////////////////////////////////
//
// Gen.cpp
//
// Scene generator.
//
// Nate Robins, January 2002
//
////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////

#include "nrArray.h"
#include "nrCmdLine.h"
#include "nrLog.h"
#include "nrLight.h"

#include "nrModel.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


////////////////////////////////////////////////////////////////////////////
// Functions
////////////////////////////////////////////////////////////////////////////

void model( const char* name )
{
    nrModel* m = nrModel::NewFromFile( name );
    if ( m == 0 )
    {
        g_Log.Write( "model(): Can't open model file \"%s\".\n", name );
        return;
    }
    
    nrVector3 mins, maxs;
    m->Dimensions( mins, maxs );
    
    nrVector3 center = ( maxs + mins ) / 2.0f;
    nrVector3 size   = ( maxs - mins );
    float scale;
    if ( size.x > size.y && size.x > size.z )
    {
        scale = 2.0f / size.x;
    }
    else if ( size.y > size.x && size.y > size.z )
    {
        scale = 2.0f / size.y;
    }
    else //if ( size.z > size.x && size.z > size.y )
    {
        scale = 2.0f / size.z;
    }
    
    printf( "# %s\n", name );
    printf( "\n\n" );
    
    printf( "view {\n" );
    printf( "  eye    < %g %g %g >\n", 0.0f, 0.0f, 3.0f );
    printf( "  lookat < %g %g %g >\n", 0.0f, 0.0f, 0.0f );
    printf( "  up     < %g %g %g >\n", 0.0f, 1.0f, 0.0f );
    printf( "}\n" );
    printf( "\n\n" );
    
    printf( "light {\n" );
    printf( "  position < %g %g %g >\n", -5.0f, 5.0f, 5.0f );
    printf( "  color < %g %g %g >\n", 1.0f, 1.0f, 0.7f );
    printf( "}\n" );
    printf( "light {\n" );
    printf( "  position < %g %g %g >\n", 5.0f, -5.0f, 5.0f );
    printf( "  color < %g %g %g >\n", 0.7f, 0.7f, 1.0f );
    printf( "}\n" );
    printf( "\n\n" );
    
    for ( int i = 0; i < m->NumGroups(); i++ )
    {
        const nrmGroup& group = m->Group( i );
        const nrmMaterial& material = group.Material();
        
        for ( int j = 0; j < group.NumTriangles(); j++ )
        {
            const nrmTriangle& t = group.Triangle( j );
            
            nrVector3 a = t.Position( 0 );
            nrVector3 b = t.Position( 1 );
            nrVector3 c = t.Position( 2 );
            
            a = a - center;
            b = b - center;
            c = c - center;
            
            a = a * scale;
            b = b * scale;
            c = c * scale;
            
            printf( "triangle {\n" );
            printf( " a < %g %g %g >\n", a.x, a.y, a.z );
            printf( " b < %g %g %g >\n", b.x, b.y, b.z );
            printf( " c < %g %g %g >\n", c.x, c.y, c.z );
            
            printf( " material {\n" );
            printf( "  diffuse < %g %g %g >\n", material.m_Diffuse.r, material.m_Diffuse.g, material.m_Diffuse.b );
            printf( "  ambient < %g %g %g >\n", material.m_Ambient.r, material.m_Ambient.g, material.m_Ambient.b );
            printf( " }\n" );
            printf( "}\n" );
        }
    }
    
    delete m;
}

////////////////////////////////////////////////////////////////////////////

void sphereflake( int depth )
{
}

////////////////////////////////////////////////////////////////////////////

void random( int n )
{
    float volume = 10.0f / n;
    float radius = nrMath::Pow( 0.75f * volume / 3.14159f, 1.0f / 3.0f );
    
    printf( "# %d random spheres\n", n );
    printf( "\n\n" );
    
    printf( "view {\n" );
    printf( "  eye    < %g %g %g >\n", 0.0f, 0.0f, 15.0f );
    printf( "  lookat < %g %g %g >\n", 0.0f, 0.0f, 0.0f );
    printf( "  up     < %g %g %g >\n", 0.0f, 1.0f, 0.0f );
    printf( "}\n" );
    printf( "\n\n" );
    
    printf( "ambient < %g %g %g >\n", 1.0f, 1.0f, 1.0f );
    printf( "\n\n" );
    
    for ( int i = 0; i < n; i++ )
    {
        float x = nrMath::Random_f( -5.0f, 5.0f );
        float y = nrMath::Random_f( -5.0f, 5.0f );
        float z = nrMath::Random_f( -5.0f, 5.0f );
        
        float r = nrMath::Random_f( 0.0f, 1.0f );
        float g = nrMath::Random_f( 0.0f, 1.0f );
        float b = nrMath::Random_f( 0.0f, 1.0f );
        
        printf( "sphere {\n" );
        printf( " center < %g %g %g >\n", x, y, z );
        printf( " radius %g\n", radius );
        printf( " color < %g %g %g >\n", r, g, b );
        printf( "}\n" );
    }
}

////////////////////////////////////////////////////////////////////////////

void colorcube( int n )
{
    float radius = ( 2.0f / n ) / 2.0f;
    
    printf( "# %dx%dx%d\n", n, n, n );
    printf( "\n\n" );
    
    printf( "view {\n" );
    printf( "  eye    < %g %g %g >\n", 0.0f, 0.0f, 3.0f );
    printf( "  lookat < %g %g %g >\n", 0.0f, 0.0f, 0.0f );
    printf( "  up     < %g %g %g >\n", 0.0f, 1.0f, 0.0f );
    printf( "}\n" );
    printf( "\n\n" );
    
    float x = ( -1.0f + radius ) + ( 2 * radius ) * 1;
    float y = ( -1.0f + radius ) + ( 2 * radius ) * ( n - 2 );
    
    printf( "light {\n" );
    printf( "  position < %g %g %g >\n", x, y, 3.0f );
    printf( "  color < %g %g %g >\n", 1.0f, 1.0f, 1.0f );
    printf( "}\n" );
    printf( "\n\n" );
    
    for ( int k = 0; k < n; k++ )
    {
        for ( int j = 0; j < n; j++ )
        {
            for ( int i = 0; i < n; i++ )
            {
                float x = ( -1.0f + radius ) + ( 2 * radius ) * i;
                float y = ( -1.0f + radius ) + ( 2 * radius ) * j;
                float z = ( -1.0f + radius ) + ( 2 * radius ) * k;
                
                float r = ( float )i / ( float )( n - 1 );
                float g = ( float )j / ( float )( n - 1 );
                float b = ( float )k / ( float )( n - 1 );
                
                printf( "sphere {\n" );
                printf( " center < %g %g %g >\n", x, y, z );
                printf( " radius %g\n", radius );
                printf( " color < %g %g %g >\n", r, g, b );
                printf( "}\n" );
            }
        }
    }
}

////////////////////////////////////////////////////////////////////////////

int main( int argc, const char** argv )
{
    char type[ 256 ];
    char args[ 256 ];
    
    nrCmdLineArg cmdlineargs[] = 
    {
        nrCmdLineArg( "-type", "<type>", 0, "type of scene to generate",     type, sizeof ( type ) ),
        nrCmdLineArg( "-args", "<args>", 0, "arguments for scene generator", args, sizeof ( args ) ),
    };
    
    nrCmdLine c( cmdlineargs, sizeof ( cmdlineargs ) / sizeof ( nrCmdLineArg ) );
    if ( ! c.Parse( argc, argv ) )
    {
        c.Usage( argv[ 0 ] );
        g_Log.Write( "    colorcube   <n>\n" );
        g_Log.Write( "       Generates a colorcube of n by n by n spheres.\n" );
        g_Log.Write( "    sphereflake <depth>\n" );
        g_Log.Write( "       Generates a sphereflake of depth specified.\n" );
        g_Log.Write( "    model <scale>\n" );
        g_Log.Write( "       Generates triangle list out of a model.\n" );
        g_Log.Write( "    random <n>\n" );
        g_Log.Write( "       Generates n random spheres.\n" );
        g_Log.Write( "\n" );
        g_Log.Write( "\n" );
        return 1;
    }
    
    nrArray<nrLight*> lights;
    
    // Depending on type, parse arguments, and call generator.
    if ( strcmp( "colorcube", type ) == 0 )
    {
        int n;
        
        if ( sscanf( args, "%d", &n ) != 1 )
        {
            g_Log.Write( "Invalid arguments for \"%s\".\n", type );
            c.Usage( argv[ 0 ] );
            return 1;
        }
        
        colorcube( n );
    }
    else if ( strcmp( "sphereflake", type ) == 0 )
    {
        int d;
        
        if ( sscanf( args, "%d", &d ) != 1 )
        {
            g_Log.Write( "Invalid arguments for \"%s\".\n", type );
            c.Usage( argv[ 0 ] );
            return 1;
        }
        
        sphereflake( d );
    }
    else if ( strcmp( "random", type ) == 0 )
    {
        int n;
        
        if ( sscanf( args, "%d", &n ) != 1 )
        {
            g_Log.Write( "Invalid arguments for \"%s\".\n", type );
            c.Usage( argv[ 0 ] );
            return 1;
        }
        
        random( n );
    }
    else if ( strcmp( "model", type ) == 0 )
    {
        char name[ 256 ];
        
        if ( sscanf( args, "%s", name ) != 1 )
        {
            g_Log.Write( "Invalid arguments for \"%s\".\n", type );
            c.Usage( argv[ 0 ] );
            return 1;
        }
        
        model( name );
    }
    else
    {
        g_Log.Write( "Invalid type \"%s\".\n", type );
        c.Usage( argv[ 0 ] );
        return 1;
    }
    
    return 0;
}

////////////////////////////////////////////////////////////////////////////
